// Randexe.cpp : Implementation of CRandexe
#include "stdafx.h"
#include "Conexe.h"
#include "Randexe.h"

void ErrorUI(HRESULT h, LPCSTR szTitle)
{
    if( h != ERROR_SUCCESS )
    {
        void *pMsgBuf;
        ::FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                         NULL,
                         h,
                         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT ),
                         (LPSTR)&pMsgBuf,
                         0,
                         NULL );

        ATLTRACE("RANDEXE: Windows error 0x%lx, %s\n", (DWORD)h, (LPSTR)pMsgBuf );         
        MessageBox( NULL, (LPSTR)pMsgBuf, szTitle, MB_ICONSTOP );
        LocalFree(pMsgBuf);
    }
}

/////////////////////////////////////////////////////////////////////////////
// CRandexe

CRandexe::CRandexe()
{
    ATLTRACE("RANDEXE: CRandexe::CRandexe\n");         
    memset(m_rsArray, 0, nMaxSessions*sizeof(RandomSessionData));
    m_nStreamIndex = 0;
}

CRandexe::~CRandexe()
{
    StopAll();
}

DWORD WINAPI RandomSessionThreadEntry(void* pv)
{
    // Need to call CoInitialize on this thread to create a single
    // threaded apartment. If you don't do this you will get the 
    // "CoInitialize has not been called." error.
    
    ErrorUI(CoInitialize(NULL), "CoInitialize error.");
    CRandexe::RandomSessionData* pS = (CRandexe::RandomSessionData*)pv;
    CRandexe* p = pS->pRandom;
    while (WaitForSingleObject(pS->m_hEvent, 0) != WAIT_OBJECT_0)
        p->Fire(pS->m_nID);
    CoUninitialize();
    return 0;
}

void CRandexe::CreateRandomSession(RandomSessionData& rs)
{
    DWORD dwThreadID = 0;
    _ASSERTE(rs.m_hEvent == NULL);
    _ASSERTE(rs.m_hThread == NULL);
    rs.pRandom = this;
    rs.m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    rs.m_hThread = CreateThread(NULL, 0, &RandomSessionThreadEntry, &rs, 0, &dwThreadID);
}

HRESULT CRandexe::Advise( IUnknown *pUnkSink, DWORD *pdwCookie )
{
    ATLTRACE("RANDEXE: CRandexe::Advise entry\n");         
    
    // Limit the number of advises in this test program.
    if( m_nStreamIndex >= 1000 )
        return CONNECT_E_CANNOTCONNECT;

    //T* pT = static_cast<T*>(this);
    IUnknown* p;
    HRESULT hRes = S_OK;
    if (pUnkSink == NULL || pdwCookie == NULL)
        return E_POINTER;
    IID iid;
    GetConnectionInterface(&iid);
    hRes = pUnkSink->QueryInterface(iid, (void**)&p);
    if (SUCCEEDED(hRes))
    {
        Lock();
        //pT->Lock();
    
        *pdwCookie = m_vec.Add(p);
        hRes = (*pdwCookie != NULL) ? S_OK : CONNECT_E_ADVISELIMIT;
        ATLTRACE("RANDEXE: CRandexe::Advise: cookie = %ld\n", *pdwCookie );         
        HRESULT hr = CoMarshalInterThreadInterfaceInStream( IID_IRandexeEvent, p, &m_StreamArray[m_nStreamIndex] );
        ErrorUI(hr, "CoMarshalInterThreadInterfaceInStream error.");
        m_nStreamIndex++;
                
        Unlock();
        //pT->Unlock();
        if (hRes != S_OK)
            p->Release();
    }
    else if (hRes == E_NOINTERFACE)
        hRes = CONNECT_E_CANNOTCONNECT;
    if (FAILED(hRes))
        *pdwCookie = 0;
    ATLTRACE("RANDEXE: CRandexe::Advise exit\n");         
    return hRes;
}

STDMETHODIMP CRandexe::Start(long* pnID)//, IRandexeEvent* pRandexeEvent)
{
    ATLTRACE("RANDEXE: CRandexe::Start\n");
    if (pnID == NULL)
        return E_POINTER;
    *pnID = 0;
    HRESULT hRes = S_OK;
    m_cs.Lock();
    for (long i=0;i<nMaxSessions;i++)
    {
        if (m_rsArray[i].m_hEvent == NULL)
        {
            m_rsArray[i].m_nID = i;
            CreateRandomSession(m_rsArray[i]);
            *pnID = i;
            break;
        }
    }
    if (i == nMaxSessions) //fell through
        hRes = E_FAIL;
    m_cs.Unlock();
    return hRes;
}


STDMETHODIMP CRandexe::Stop(long nID)
{
    HRESULT hRes = S_OK;
    m_cs.Lock();
    if (m_rsArray[nID].m_hEvent != NULL)
    {
        SetEvent(m_rsArray[nID].m_hEvent);
        WaitForSingleObject(m_rsArray[nID].m_hThread, INFINITE);
        CloseHandle(m_rsArray[nID].m_hThread);
        memset(&m_rsArray[nID], 0, sizeof(RandomSessionData));
    }
    else
        hRes = E_INVALIDARG;
    m_cs.Unlock();
    return hRes;
}

STDMETHODIMP CRandexe::StopAll()
{
    m_cs.Lock();
    for (long i=0;i<nMaxSessions;i++)
    {
        if (m_rsArray[i].m_hEvent != NULL)
        {
            SetEvent(m_rsArray[i].m_hEvent);
            WaitForSingleObject(m_rsArray[i].m_hThread, INFINITE);
            CloseHandle(m_rsArray[i].m_hThread);
            memset(&m_rsArray[i], 0, sizeof(RandomSessionData));
        }
    }
    m_cs.Unlock();
    return S_OK;
}

// broadcast to all the objects
HRESULT CRandexe::Fire(long nID)
{
    Lock();
    HRESULT hr = S_OK;
    for( int i = 0; i < m_nStreamIndex; i++ )
    {
        CComPtr<IStream> pStream;
        hr = m_StreamArray[i]->Clone( &pStream );

        IRandexeEvent *pI;
        hr = CoUnmarshalInterface( pStream, IID_IRandexeEvent, (void**)&pI );

        if(FAILED(hr))
        {
            void *pMsgBuf;
            ::FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                             NULL,
                             hr,
                             MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT ),
                             (LPSTR)&pMsgBuf,
                             0,
                             NULL );

            ATLTRACE("RANDEXE: Windows error 0x%lx, %s\n", (DWORD)hr, (LPSTR)pMsgBuf );         
            LocalFree(pMsgBuf);
        }
        hr = pI->Fire(nID);
    }           
    Unlock();
    return hr;
}












